@@ -1,7 +1,7 @@
 package App::shcompgen;
 
-our $DATE = '2016-07-12'; # DATE
-our $VERSION = '0.19'; # VERSION
+our $DATE = '2016-10-21'; # DATE
+our $VERSION = '0.20'; # VERSION
 
 use 5.010001;
 use strict;
@@ -22,18 +22,34 @@ $SPEC{':package'} = {
     summary => 'Generate shell completion scripts',
 };
 
-our @supported_shells = qw(bash); # XXX fish tcsh zsh
-our %common_args = (
+my $_complete_prog = sub {
+    require Complete::File;
+    require Complete::Program;
+
+    my %args = @_;
+    my $word = $args{word} // '';
+    if ($word =~ m!/!) {
+        # user might want to mention a program file (e.g. ./foo)
+        return {
+            words => Complete::File::complete_file(
+                word=>$word, ci=>1, filter=>'d|rxf'),
+            path_sep => '/',
+        };
+    } else {
+        # or user might want to mention a program in PATH
+        Complete::Program::complete_program(word=>$word, ci=>1);
+    }
+};
+
+our @supported_shells = qw(bash fish); # XXX tcsh zsh
+our %shell_arg = (
     shell => {
-        summary => 'Override autodetection and select shell manually',
+        summary => 'Override guessing and select shell manually',
         schema => ['str*', {in=>\@supported_shells}],
-        description => <<'_',
-
-The default is to look at your SHELL environment variable value. If it is
-undefined, the default is `bash`.
-
-_
     },
+);
+our %common_args = (
+    %shell_arg,
     global => {
         summary => 'Use global completions directory',
         schema => ['bool*'],
@@ -68,15 +84,15 @@ _
         schema  => ['array*', of => 'str*'],
     },
 
-    #fish_global_dir => {
-    #    summary => 'Directory to put completions scripts',
-    #    schema  => ['array*', of => 'str*'],
-    #    default => ['/usr/share/fish/completions', '/etc/fish/completions'],
-    #},
-    #fish_per_user_dir => {
-    #    summary => 'Directory to put completions scripts',
-    #    schema  => ['array*', of => 'str*'],
-    #},
+    fish_global_dir => {
+        summary => 'Directory to put completions scripts',
+        schema  => ['array*', of => 'str*'],
+        default => ['/usr/share/fish/completions', '/etc/fish/completions'],
+    },
+    fish_per_user_dir => {
+        summary => 'Directory to put completions scripts',
+        schema  => ['array*', of => 'str*'],
+    },
 
     #tcsh_global_dir => {
     #    summary => 'Directory to put completions scripts',
@@ -117,8 +133,13 @@ sub _all_exec_in_PATH {
 sub _set_args_defaults {
     my $args = shift;
 
-    if (!$args->{shell}) { ($args->{shell} = $ENV{SHELL} // '') =~ s!.+/!! }
-    if (!$args->{shell}) { $args->{shell} = 'bash' }
+    if (!$args->{shell}) {
+        require Shell::Guess;
+        my $sh = Shell::Guess->running_shell;
+        my $n = $sh->{name};
+        $n = "zsh" if $n eq 'z';
+        $args->{shell} = $n;
+    }
     unless ($args->{shell} ~~ @supported_shells) {
         return [412, "Unsupported shell '$args->{shell}'"];
     }
@@ -131,10 +152,11 @@ sub _set_args_defaults {
     $args->{fish_global_dir}   //= ['/usr/share/fish/completions',
                                     '/etc/fish/completions'];
     $args->{fish_per_user_dir} //= ["$ENV{HOME}/.config/fish/completions"];
-    $args->{tcsh_global_dir}   //= [];
-    $args->{tcsh_per_user_dir} //= ["$ENV{HOME}/.config/tcsh/completions"];
-    $args->{zsh_global_dir}    //= [];
-    $args->{zsh_per_user_dir}  //= ["$ENV{HOME}/.config/zsh/completions"];
+    #$args->{tcsh_global_dir}   //= [];
+    #$args->{tcsh_per_user_dir} //= ["$ENV{HOME}/.config/tcsh/completions"];
+    #$args->{zsh_global_dir}    //= [];
+    #$args->{zsh_per_user_dir}  //= ["$ENV{HOME}/.config/zsh/completions"];
+    [200];
 }
 
 sub _gen_completion_script {
@@ -180,14 +202,44 @@ complete -F _|."$prog $qprog".q|
         } else {
             $script = "complete -C $qcomp $qprog";
         }
+    } elsif ($shell eq 'fish') {
+        require File::Which;
+        my $path = File::Which::which($comp);
+        my $type = $detres->[3]{'func.completer_type'};
+        if ($type eq 'Getopt::Long' || $type eq 'Getopt::Long::Complete') {
+            require Complete::Fish::Gen::FromGetoptLong;
+            my $gen_res = Complete::Fish::Gen::FromGetoptLong::gen_fish_complete_from_getopt_long_script(
+                filename => $path,
+                cmdname => $prog,
+                compname => $comp,
+            );
+            if ($gen_res->[0] != 200) {
+                $log->errorf("Can't generate fish completion script for '%s': %s", $path, $gen_res);
+                $script = "# Can't generate fish completion script for '$path': $gen_res->[0] - $gen_res->[1]\n";
+                goto L1;
+            }
+            $script = $gen_res->[2];
+        } elsif ($type eq 'Perinci::CmdLine') {
+            require Complete::Fish::Gen::FromPerinciCmdLine;
+            my $gen_res = Complete::Fish::Gen::FromPerinciCmdLine::gen_fish_complete_from_perinci_cmdline_script(
+                filename => $path,
+                cmdname => $prog,
+                compname => $comp,
+            );
+            if ($gen_res->[0] != 200) {
+                $log->errorf("Can't generate fish completion script for '%s': %s", $path, $gen_res);
+                $script = "# Can't generate fish completion script for '$path': $gen_res->[0] - $gen_res->[1]\n";
+                goto L1;
+            }
+            $script = $gen_res->[2];
+        } else {
+            $script = "# TODO for type=$type\n";
+        }
     } else {
-        die "Sorry, shells other than bash are not supported yet";
+        die "Sorry, shells other than bash/fish are not supported yet";
     }
 
-    # fish
-    # - completer_command -> check completer command if it's pericmd or
-    # glcomp or glsubc
-
+  L1:
     $script = "# FRAGMENT id=shcompgen-header note=".
         ($detres->[3]{'func.note'} // ''). "\n$script\n";
 
@@ -239,10 +291,12 @@ sub _completion_script_path {
     $path;
 }
 
-# XXX plugin based
+# detect whether we can generate completion script for a program, under a given
+# shell
 sub _detect_prog {
     my %args = @_;
 
+    my $shell    = $args{shell};
     my $prog     = $args{prog};
     my $progpath = $args{progpath};
 
@@ -257,6 +311,8 @@ sub _detect_prog {
     seek $fh, 0, 0;
     my $content = do { local $/; ~~<$fh> };
 
+    my %extrametas;
+
   DETECT:
     {
         if ($content =~
@@ -270,34 +326,123 @@ sub _detect_prog {
                 $args =~ s/"\z//;
                 $args =~ s/\\(.)/$1/g;
             }
+            if ($shell eq 'fish') {
+                # for fish, we need to make sure first that we can extract
+                # cmdline-options from the completer program
+                require File::Which;
+                my $cmdpath = File::Which::which($cmd);
+                my $cmddet_res = _detect_prog(
+                    prog => $cmd,
+                    progpath => $cmdpath,
+                    shell => $shell,
+                );
+                my $reason;
+                {
+                    if ($cmddet_res->[0] != 200) {
+                        $reason = "$cmddet_res->[0] - $cmddet_res->[1]";
+                        last;
+                    }
+                    if (!$cmddet_res->[2]) {
+                        $reason = "'$cmd' is not supported by shcompgen";
+                    }
+                    if ($cmddet_res->[3]{'completer_command'} &&
+                            $cmddet_res->[3]{'completer_command'} ne $cmd) {
+                        $reason = "multiple indirection of completer is currently not supported";
+                        last;
+                    }
+                    if ($cmddet_res->[3]{'func.completer_type'} && $cmddet_res->[3]{'func.completer_type'} !~
+                            /^(Getopt::Long(::Complete)?|Perinci::CmdLine(?:::\w+)?)$/) {
+                        $reason = "currently only Getopt::Long-/Getopt::Long::Complete-/Perinci::CmdLine-based scripts are supported";
+                        last;
+                    }
+                    $extrametas{'func.completer_type'} =
+                        $cmddet_res->[3]{'func.completer_type'};
+                }
+                return [200, "Program '$prog' is completed by ".
+                            "'$cmd' (hint(command)), but we don't support ".
+                            "creating completion script with '$cmd'".
+                            ": $reason", 0]
+                    if $reason;
+            }
             return [200, "OK", 1, {
                 "func.completer_command" => $cmd,
                 "func.completer_command_args" => $args,
                 "func.note" => "hint(command)",
+                %extrametas,
             }];
         }
-        if($content =~
+        if(!$args{_skip_completer_hint} &&
+               $content =~
                /^\s*# FRAGMENT id=shcompgen-hint completer=1 for=(.+?)\s*$/m
                && $content !~ /^\s*# FRAGMENT id=shcompgen-nohint\s*$/m) {
             my $completee = $1;
             return [400, "completee specified in '$progpath' is not a valid ".
                         "program name: $completee"]
                 unless $completee =~ $re_progname;
+            if ($shell eq 'fish') {
+                # for fish, we need to make sure first that we can extract
+                # cmdline-options from the completer program
+                require File::Which;
+                my $det_res2 = _detect_prog(
+                    prog => $prog,
+                    progpath => $progpath,
+                    shell => $shell,
+                    _skip_completer_hint=>1,
+                );
+                my $reason;
+                {
+                    if ($det_res2->[0] != 200) {
+                        $reason = "$det_res2->[0] - $det_res2->[1]";
+                        last;
+                    }
+                    if (!$det_res2->[2]) {
+                        $reason = "'$prog' is not supported by shcompgen";
+                    }
+                    if ($det_res2->[3]{'completer_command'} &&
+                            $det_res2->[3]{'completer_command'} ne $prog) {
+                        $reason = "multiple indirection of completer is currently not supported";
+                        last;
+                    }
+                    if ($det_res2->[3]{'func.completer_type'} && $det_res2->[3]{'func.completer_type'} !~
+                            /^(Getopt::Long(::Complete|::Subcommand)?|Perinci::CmdLine(?:::\w+)?)$/) {
+                        $reason = "currently only Getopt::Long-/Getopt::Long::Complete-/Perinci::CmdLine-based scripts are supported";
+                        last;
+                    }
+                    $extrametas{'func.completer_type'} = $det_res2->[3]{'func.completer_type'};
+                }
+                return [200, "Program '$prog' is a completer for ".
+                            "'$completee' (hint(completer)), but we don't support ".
+                            "creating completion script with '$prog'".
+                            ": $reason", 0]
+                    if $reason;
+            }
             return [200, "OK", 1, {
                 "func.completer_command" => $prog,
                 "func.completee" => $completee,
                 "func.note"=>"hint(completer)",
+                %extrametas,
             }];
         }
+
         if ($is_perl_script &&
                 # regex split here because i found a pathological case of very
                 # long matching time againt 'rsybak' datapacked script (~ 4M)
-                $content =~ /^\s*(use|require)\s+Getopt::Long::Complete\b/m ||
-                $content =~ /^\s*(use|require)\s+Getopt::Long::Subcommand\b/m) {
+                $content =~ /^\s*((?:use|require)\s+(Getopt::Long::Complete))\b/m ||
+                $shell ne 'fish' &&
+                $content =~ /^\s*((?:use|require)\s+(Getopt::Long::Subcommand))\b/m) {
             return [200, "OK", 1, {
                 "func.completer_command"=> $prog,
                 "func.completer_type"=> $2,
-                "func.note"=>$2,
+                "func.note"=>"perl use/require statement: $1",
+            }];
+        }
+
+        if ($shell eq 'fish' && $is_perl_script &&
+                $content =~ /^\s*((?:use|require)\s+(Getopt::Long))\b/m) {
+            return [200, "OK", 1, {
+                "func.completer_command" => $prog,
+                "func.completer_type" => $2,
+                "func.note"=>"perl use/require statement: $1",
             }];
         }
 
@@ -317,6 +462,7 @@ sub _detect_prog {
                 "func.completer_command"=> $prog,
                 "func.completer_type"=> "Perinci::CmdLine",
                 "func.note"=>"detected using Perinci::CmdLine::Util",
+                "func.pericmd_detect_result" => $det_res,
             }];
         }
     }
@@ -327,7 +473,8 @@ sub _generate_or_remove {
     my $which0 = shift;
     my %args = @_;
 
-    _set_args_defaults(\%args);
+    my $setdef_res = _set_args_defaults(\%args);
+    return $setdef_res unless $setdef_res->[0] == 200;
 
     # to avoid writing a file and then removing the file again in the same run
     my %written_files;
@@ -358,7 +505,7 @@ sub _generate_or_remove {
 
         my $which = $which0;
         if ($which eq 'generate') {
-            my $detres = _detect_prog(prog=>$prog, progpath=>$progpath);
+            my $detres = _detect_prog(prog=>$prog, progpath=>$progpath, shell=>$args{shell});
             if ($detres->[0] != 200) {
                 $log->errorf("Can't detect '%s': %s", $prog, $detres->[1]);
                 $envres->add_result($detres->[0], $detres->[1],
@@ -380,6 +527,11 @@ sub _generate_or_remove {
             my $comppath = _completion_script_path(
                 %args, prog => $prog, detect_res => $detres);
 
+            if ($args{stdout}) {
+                print $script;
+                next PROG;
+            }
+
             if (-f $comppath) {
                 if (!$args{replace}) {
                     $log->infof("Not replacing completion script for $prog in '$comppath' (use --replace to replace)");
@@ -435,6 +587,55 @@ sub _generate_or_remove {
     $envres->as_struct;
 }
 
+$SPEC{guess_shell} = {
+    v => 1.1,
+    summary => 'Guess running shell',
+    args => {
+    },
+};
+sub guess_shell {
+    my %args = @_;
+
+    my $setdef_res = _set_args_defaults(\%args);
+    return $setdef_res unless $setdef_res->[0] == 200;
+
+    [200, "OK", $args{shell}];
+}
+
+$SPEC{detect_prog} = {
+    v => 1.1,
+    summary => "Detect a program",
+    args => {
+        %shell_arg,
+        prog => {
+            schema => 'str*',
+            completion => $_complete_prog,
+            req => 1,
+            pos => 0,
+        },
+    },
+    'cmdline.default_format' => 'json',
+};
+sub detect_prog {
+    require File::Which;
+
+    my %args = @_;
+
+    _set_args_defaults(\%args);
+
+    my $progname = $args{prog};
+    my $progpath = File::Which::which($progname);
+
+    return [404, "No such program '$progname'"] unless $progpath;
+    $progname =~ s!.+/!!;
+
+    _detect_prog(
+        prog => $progname,
+        progpath => $progpath,
+        shell => $args{shell},
+    );
+}
+
 $SPEC{init} = {
     v => 1.1,
     summary => 'Initialize shcompgen',
@@ -451,7 +652,9 @@ _
 sub init {
     my %args = @_;
 
-    _set_args_defaults(\%args);
+    my $setdef_res = _set_args_defaults(\%args);
+    return $setdef_res unless $setdef_res->[0] == 200;
+
     my $shell = $args{shell};
     my $global = $args{global};
 
@@ -505,6 +708,9 @@ _
         } else {
             $init_script_path = "$ENV{HOME}/.config/shcompgen.bashrc";
         }
+    } elsif ($shell eq 'fish') {
+        # nothing to do, should be ready
+        return [200, "OK"];
     } else {
         return [412, "Shell '$shell' not yet supported"];
     }
@@ -544,24 +750,7 @@ Can contain path (e.g. `../foo`) or a plain word (`foo`) in which case will be
 searched from PATH.
 
 _
-            element_completion => sub {
-                require Complete::File;
-                require Complete::Program;
-
-                my %args = @_;
-                my $word = $args{word} // '';
-                if ($word =~ m!/!) {
-                    # user might want to mention a program file (e.g. ./foo)
-                    return {
-                        words => Complete::File::complete_file(
-                            word=>$word, ci=>1, filter=>'d|rxf'),
-                        path_sep => '/',
-                    };
-                } else {
-                    # or user might want to mention a program in PATH
-                    Complete::Program::complete_program(word=>$word, ci=>1);
-                }
-            },
+            element_completion => $_complete_prog,
         },
         replace => {
             summary => 'Replace existing script',
@@ -584,6 +773,10 @@ enabled, however, such existing completion script will be removed.
 
 _
         },
+        stdout => {
+            summary => 'Output completion script to STDOUT',
+            schema => ['bool', is=>1],
+        },
     },
 };
 sub generate {
@@ -599,15 +792,18 @@ $SPEC{list} = {
         %common_args,
         detail => {
             schema => 'bool',
+            cmdline_aliases => {l=>{}},
         },
     },
 };
 sub list {
     my %args = @_;
 
-    _set_args_defaults(\%args);
+    my $setdef_res = _set_args_defaults(\%args);
+    return $setdef_res unless $setdef_res->[0] == 200;
 
     my @res;
+    my $resmeta = {};
     my $dirs = _completion_scripts_dirs(%args);
     for my $dir (@$dirs) {
         $log->debugf("Opening dir %s ...", $dir);
@@ -646,7 +842,9 @@ sub list {
         }
     } # for $dir
 
-    [200, "OK", \@res];
+    $resmeta->{'table.fields'} = [qw/prog path note/] if $args{detail};
+
+    [200, "OK", \@res, $resmeta];
 }
 
 $SPEC{remove} = {
@@ -701,11 +899,41 @@ App::shcompgen - Generate shell completion scripts
 
 =head1 VERSION
 
-This document describes version 0.19 of App::shcompgen (from Perl distribution App-shcompgen), released on 2016-07-12.
+This document describes version 0.20 of App::shcompgen (from Perl distribution App-shcompgen), released on 2016-10-21.
 
 =head1 FUNCTIONS
 
 
+=head2 detect_prog(%args) -> [status, msg, result, meta]
+
+Detect a program.
+
+This function is not exported.
+
+Arguments ('*' denotes required arguments):
+
+=over 4
+
+=item * B<prog>* => I<str>
+
+=item * B<shell> => I<str>
+
+Override guessing and select shell manually.
+
+=back
+
+Returns an enveloped result (an array).
+
+First element (status) is an integer containing HTTP status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+(msg) is a string containing error message, or 'OK' if status is
+200. Third element (result) is optional, the actual result. Fourth
+element (meta) is called result metadata and is optional, a hash
+that contains extra information.
+
+Return value:  (any)
+
+
 =head2 generate(%args) -> [status, msg, result, meta]
 
 Generate shell completion scripts for detectable programs.
@@ -724,6 +952,14 @@ Directory to put completions scripts.
 
 Directory to put completions scripts.
 
+=item * B<fish_global_dir> => I<array[str]> (default: ["/usr/share/fish/completions","/etc/fish/completions"])
+
+Directory to put completions scripts.
+
+=item * B<fish_per_user_dir> => I<array[str]>
+
+Directory to put completions scripts.
+
 =item * B<global> => I<bool>
 
 Use global completions directory.
@@ -759,10 +995,11 @@ The default behavior is to skip if an existing completion script exists.
 
 =item * B<shell> => I<str>
 
-Override autodetection and select shell manually.
+Override guessing and select shell manually.
 
-The default is to look at your SHELL environment variable value. If it is
-undefined, the default is C<bash>.
+=item * B<stdout> => I<bool>
+
+Output completion script to STDOUT.
 
 =back
 
@@ -778,6 +1015,26 @@ that contains extra information.
 Return value:  (any)
 
 
+=head2 guess_shell() -> [status, msg, result, meta]
+
+Guess running shell.
+
+This function is not exported.
+
+No arguments.
+
+Returns an enveloped result (an array).
+
+First element (status) is an integer containing HTTP status code
+(200 means OK, 4xx caller error, 5xx function error). Second element
+(msg) is a string containing error message, or 'OK' if status is
+200. Third element (result) is optional, the actual result. Fourth
+element (meta) is called result metadata and is optional, a hash
+that contains extra information.
+
+Return value:  (any)
+
+
 =head2 init(%args) -> [status, msg, result, meta]
 
 Initialize shcompgen.
@@ -799,6 +1056,14 @@ Directory to put completions scripts.
 
 Directory to put completions scripts.
 
+=item * B<fish_global_dir> => I<array[str]> (default: ["/usr/share/fish/completions","/etc/fish/completions"])
+
+Directory to put completions scripts.
+
+=item * B<fish_per_user_dir> => I<array[str]>
+
+Directory to put completions scripts.
+
 =item * B<global> => I<bool>
 
 Use global completions directory.
@@ -813,10 +1078,7 @@ that and manually select which.
 
 =item * B<shell> => I<str>
 
-Override autodetection and select shell manually.
-
-The default is to look at your SHELL environment variable value. If it is
-undefined, the default is C<bash>.
+Override guessing and select shell manually.
 
 =back
 
@@ -852,6 +1114,14 @@ Directory to put completions scripts.
 
 =item * B<detail> => I<bool>
 
+=item * B<fish_global_dir> => I<array[str]> (default: ["/usr/share/fish/completions","/etc/fish/completions"])
+
+Directory to put completions scripts.
+
+=item * B<fish_per_user_dir> => I<array[str]>
+
+Directory to put completions scripts.
+
 =item * B<global> => I<bool>
 
 Use global completions directory.
@@ -866,10 +1136,7 @@ that and manually select which.
 
 =item * B<shell> => I<str>
 
-Override autodetection and select shell manually.
-
-The default is to look at your SHELL environment variable value. If it is
-undefined, the default is C<bash>.
+Override guessing and select shell manually.
 
 =back
 
@@ -903,6 +1170,14 @@ Directory to put completions scripts.
 
 Directory to put completions scripts.
 
+=item * B<fish_global_dir> => I<array[str]> (default: ["/usr/share/fish/completions","/etc/fish/completions"])
+
+Directory to put completions scripts.
+
+=item * B<fish_per_user_dir> => I<array[str]>
+
+Directory to put completions scripts.
+
 =item * B<global> => I<bool>
 
 Use global completions directory.
@@ -924,10 +1199,7 @@ searched from PATH.
 
 =item * B<shell> => I<str>
 
-Override autodetection and select shell manually.
-
-The default is to look at your SHELL environment variable value. If it is
-undefined, the default is C<bash>.
+Override guessing and select shell manually.
 
 =back
 
